#ifndef Schrift_H
#define Schrift_H

#include <functional>

#include "Critical.h"
#include "Punkt.h"
#include "ReferenceCounter.h"

namespace Framework
{
    class Bild;      //! Bild.h
    class Text;      //! Text.h
    class Buchstabe; //! aus dieser Datei
    class Alphabet;  //! aus dieser Datei
    class Schrift;   //! aus dieser Datei

    //! Speichert die Alphawerte eines Zeichens einer bestimmten Schrift
    //! Die anderen Farbwerte werden durch die Schriftfarbe bestimmt. Daher nur
    //! die Alpha werte.
    class Buchstabe : public virtual ReferenceCounter
    {
    private:
        Punkt size;
        unsigned char* alpha;
        int schriftSize;

    public:
        //! Erstellt ein neues leeres Zeichnung
        DLLEXPORT Buchstabe();
        //! Lscht das Zeichnung
        DLLEXPORT ~Buchstabe();
        //! Erstellt einen neuen Buchstaben mit bestimmter Gre
        //! \param size Die Gre des Buchstabens in Pixel
        DLLEXPORT void NeuBuchstabe(Punkt& size);
        //! Setzt den Alphawert eines bestimmten Pixels
        //! \param pos Die position des Pixels
        //! \param alpha Der Wert des Pixels
        DLLEXPORT void setPixel(Punkt& pos, unsigned char alpha);
        //! Setzt den Alphawert eines bestimmten Pixels
        //! \param x die x Position des Pixels
        //! \param y die y Position des Pixels
        //! \param alpha Der Wert des Pixels
        DLLEXPORT void setPixel(int x, int y, unsigned char alpha);
        //! Setzt den Alphawert eines bestimmten Pixels
        //! \param i der Index des Pixels. Er berechnet sich durch x + y *
        //! breite und geht von 0 bis breite * hhe - 1 \param alpha Der Wert
        //! des Pixels
        DLLEXPORT void setPixel(int i, unsigned char alpha);
        //! Setzt die Schriftgre, zu der der Buchstabe gehrt
        //! \param sg Die Schriftgre des Buchstabens.
        DLLEXPORT void setSchriftSize(int sg);
        //! Gibt die Schriftgre zurck, zu der der Buchstabe gehrt
        DLLEXPORT int getSchriftSize() const;
        //! Gibt die alpha Werte des Buchstabens als array zurck wobei die
        //! werte Zeilenweise hintereinander stehen
        DLLEXPORT unsigned char* getBuff() const;
        //! Gibt die Gre des Buchstabens in Pixeln nicht skalliert zurck.
        DLLEXPORT const Punkt& getSize() const;
        //! Gibt die Breite des Buchstabens in Pixeln zurck
        DLLEXPORT int getBreite() const;
        //! Gibt die Hhe des Buchstabens in Pixeln zurck
        DLLEXPORT int getHeight() const;
    };

    //! Speichert alle Buchstaben der selben Schriftgre.
    //! Wird von der Schrift klasse verwendet
    class Alphabet : public virtual ReferenceCounter
    {
    private:
        Buchstabe** zeichen;
        int schriftSize;

    public:
        //! Erzeugt ein leeres Zeichnung
        DLLEXPORT Alphabet();
        //! Lscht ein Zeichnung
        DLLEXPORT ~Alphabet();
        //! Lscht alle gespeicherten Zeichen
        DLLEXPORT void NeuAlphabet();
        //! Fgt dem Alphabet einen Buchstaben hinzu
        //! Wenn der hinzugefgte Buchstabe bereits existiert wird er
        //! berschrieben \param i Der ASCII code des Buchstaben, der
        //! hinzugefgt werden soll \param buchstabe Der Buchstabe, der
        //! hinzugefgt wird
        DLLEXPORT void setBuchstabe(unsigned char i, Buchstabe* buchstabe);
        //! Setzt die Schriftgre des Alphabets und die der gespeicherten
        //! buchstaben \param gr Die Schriftgre des Alphabets
        DLLEXPORT void setSchriftSize(int gr);
        //! Gibt den gespeicherten Buchstaben zu einem bestimmten ASCII Zeichen
        //! zurck \param i Der ASCII code des Zeichens \return Ein Zeiger zu
        //! dem Buchstaben mit erhhtem Reference Counter
        DLLEXPORT Buchstabe* getBuchstabe(unsigned char i) const;
        //! Gibt den gespeicherten Buchstaben zu einem bestimmten ASCII Zeichen
        //! zurck \param i Der ASCII code des Zeichens \return Ein Zeiger zu
        //! dem Buchstaben ohne erhhtem Reference Counter
        DLLEXPORT Buchstabe* zBuchstabe(unsigned char i) const;
        //! Prft, ob zu einem bestimmten ASCII code ein Zeichen vorhanden ist
        //! \param b Der zu prfende ASCII code
        //! \return (true), wenn ein Zeichen zu dem Code gefunden wurde. (false)
        //! sonnst
        DLLEXPORT bool hatBuchstabe(unsigned char b) const;
        //! Gibt die Schriftgre zurck, deren Zeichen in diesem Alphabet
        //! gespeichert werden
        DLLEXPORT int getSchriftSize() const;
    };

    //! Specihert eine Liste von Alphabeten mit verschiedener Schriftgre.
    //! Wird von der Schrift Klasse verwendet, um alle Zeichen der Schriftgre
    //! nach sortiert zu speichern.
    class AlphabetArray
    {
    private:
        Alphabet* alphabets[256];

    public:
        //! Erzeugt eine neue Liste
        DLLEXPORT AlphabetArray();

        //! Fgt der Liste ein Alphabet hinzu
        //! Wenn bereits ein Alphabet mit der selben Schriftgre existiert,
        //! wird das Alphabet nicht hinzugefgt \param alphabet Das Alphabet,
        //! welches hinzugefgt werden soll \return (true), wenn das Alphabet
        //! hinzugefgt wurde. (false) sonnst.
        DLLEXPORT bool addAlphabet(Alphabet* alphabet);
        //! Lscht ein Alphabet bestimmter Schriftgre aus der Liste
        //! \param sg Die Schriftgre des Alphabets, welches entfernt werden
        //! soll \return (true), wenn ein Alphabet entfernt wurde. (false)
        //! sonnst
        DLLEXPORT bool removeAlphabet(unsigned char sg);

        //! Gibt ein bestimmtes Alphabet mit erhhtem Reference Counter zurck
        //! \param sg Die Schriftgre, dessen Alphabet gesucht werden soll
        //! \return (0), fals kein passendes Alphabet gefunden wurde
        DLLEXPORT Alphabet* getAlphabet(unsigned char sg) const;
        //! Gibt ein bestimmtes Alphabet ohne erhhtem Reference Counter zurck
        //! \param sg Die Schriftgre, dessen Alphabet gesucht werden soll
        //! \return (0), fals kein passendes Alphabet gefunden wurde
        DLLEXPORT Alphabet* zAlphabet(unsigned char sg) const;
    };

    //! Speichert alle Buchstaben einer Schrift in verschiedenen Schriftgren
    class Schrift : public virtual ReferenceCounter
    {
    private:
        unsigned char alphabetAnzahl;
        AlphabetArray* alphabet;

    public:
        //! Erzeugt eine leere Schrift
        DLLEXPORT Schrift();
        //! Lscht ein Zeichnung
        DLLEXPORT ~Schrift();
        //! Fgt der Schrift ein Alphabet hinzu. Sollte bereits ein Alphabet der
        //! selben Schriftgre existieren, wird das Alphabet nicht hinzugefgt
        //! \param alphabet Das Alphabet, welches hinzugefgt werden soll
        //! \return (true), wenn das Alphabet hinzugefgt wurde. (false) sonnst
        DLLEXPORT bool addAlphabet(Alphabet* alphabet);
        //! Lscht ein bestimmtes Alphabet aus der Schrift
        //! \param sg Die Schriftgre, deren Alphabet entfernt werden soll
        DLLEXPORT void removeAlphabet(unsigned char sg);
        //! Gibt ein bestimmtes Alphabet mit erhhtem Reference Counter zurck
        //! \param sg Die Schriftgre, dessen Alphabet gesucht werden soll
        //! \return (0), fals kein passendes Alphabet gefunden wurde
        DLLEXPORT Alphabet* getAlphabet(unsigned char sg) const;
        //! Gibt ein bestimmtes Alphabet ohne erhhtem Reference Counter zurck
        //! \param sg Die Schriftgre, dessen Alphabet gesucht werden soll
        //! \return (0), fals kein passendes Alphabet gefunden wurde
        DLLEXPORT Alphabet* zAlphabet(unsigned char sg) const;
        //! Gibt zurck, wie viele Alphabete (und damit Schriftgren) in der
        //! Schrift enthalten sind
        DLLEXPORT unsigned char getAlphabetAnzahl() const;
    };

    class TextRenderer : public virtual ReferenceCounter
    {
    protected:
        Schrift* s;
        int schriftSize;
        int zeilenAbstand;
        int zeichenAbstand;
        int charWidths[256];
        int charHeights[256];

    public:
        DLLEXPORT TextRenderer();
        DLLEXPORT TextRenderer(Schrift* schrift);
        DLLEXPORT virtual ~TextRenderer();
        DLLEXPORT void setSchriftZ(Schrift* schrift);
        DLLEXPORT Schrift* getSchrift();
        DLLEXPORT Schrift* zSchrift();
        //! Setzt die Schriftgre, in der gezeichnet werden soll. Die Schrift
        //! whlt automatisch das passende Alphabet zum Zeichnen \param sg Die
        //! Schriftgre
        DLLEXPORT void setSchriftSize(int sg);
        //! Setzt den Zeilenabstand, der zum zeichnen verwendet werden soll
        //! \param za Der Zeilenabstand zum unteren Ende der darber liegenden
        //! zeile in Pixeln
        DLLEXPORT void setZeilenAbstand(int za);
        //! Setzt den Zeichenabstand, der zum zeichnen verwendet werden soll
        //! \param za Der Zeichenabstand zum unteren Ende der darber liegenden
        //! zeile in Pixeln
        DLLEXPORT void setZeichenAbstand(int za);
        //! Fgt Zeilenumbrche in den Text ein, so dass er bei einer
        //! vorgegebenen Breite follstndig angezeigt wird \param zText Der
        //! text, in den die Zeilenumbrche eingefgt werden sollen \param
        //! maxBreite Die Breite in Pixeln auf der der Text follstndig
        //! angezeigt werden soll
        DLLEXPORT virtual void textFormatieren(Text* zText, int maxBreite);
        //! Zeichnet einen Bestimmten Text mit Cursor und einfrbung auf ein
        //! Bild Nutze (setPosition) und (setDrawSchriftGre) um die Position
        //! und die Gre zu verndern \param x x position des ersten zeichens
        //! \param y y position des ersten zeichens
        //! \param txt Der Text, der gezeichnet werden soll
        //! \param zRObj Das Bild, auf das gezeichnet werden soll
        //! \param cpos Die position des Cursors im Text
        //! \param cf Die Farbe des Cursors
        //! \param fbeg Die Position des Zeichens im Text, wo die Einfrbung
        //! beginnen soll. Der Text wird von dort bis zur Cursorposition
        //! eingefrbt \param ff Die Hintergrund Farbe des eingefrbten Textes
        //! \param f Eine Funktion die fr jeden Buchstaben aufgerufen wird und
        //! seine Farbe zurckgibt
        DLLEXPORT virtual void renderText(int x,
            int y,
            const char* txt,
            Bild& zRObj,
            std::function<int(int, int, int)> f,
            int cpos = -1,
            int cf = 0,
            int fbeg = -1,
            int ff = 0);
        //! Zeichnet einen Bestimmten Text mit Cursor und einfrbung auf ein
        //! Bild Nutze (setPosition) und (setDrawSchriftGre) um die Position
        //! und die Gre zu verndern \param x x position des ersten zeichens
        //! \param y y position des ersten zeichens
        //! \param txt Der Text, der gezeichnet werden soll
        //! \param zRObj Das Bild, auf das gezeichnet werden soll
        //! \param cpos Die position des Cursors im Text
        //! \param cf Die Farbe des Cursors
        //! \param fbeg Die Position des Zeichens im Text, wo die Einfrbung
        //! beginnen soll. Der Text wird von dort bis zur Cursorposition
        //! eingefrbt \param ff Die Hintergrund Farbe des eingefrbten Textes
        //! \param f Die Farbe, in der der Text gezeichnet werden soll
        DLLEXPORT virtual void renderText(int x,
            int y,
            const char* txt,
            Bild& zRObj,
            int f,
            int cpos = -1,
            int cf = 0,
            int fbeg = -1,
            int ff = 0);
        //! Zeichnet einen Bestimmten Buchstaben mit einfrbung auf ein Bild
        //! Nutze (setPosition) und (setDrawSchriftGre) um die Position und
        //! die Gre zu verndern \param x x position des ersten zeichens
        //! \param y y position des ersten zeichens
        //! \param txt Der Text, der gezeichnet werden soll
        //! \param zRObj Das Bild, auf das gezeichnet werden soll
        //! \param color Die Farbe, in der der Text gezeichnet werden soll
        //! \param underlined 1, falls der Text unterstrichen sein soll
        //! \param selected 1, falls das zeichen eingefrbt sein soll
        //! \param selectedBackgroundColor Die Hintergrund Farbe des
        //! eingefrbten Textes
        DLLEXPORT virtual void renderChar(int& x,
            int y,
            char c,
            Bild& zRObj,
            int color,
            bool underlined = 0,
            bool selected = 0,
            int selectedBackgroundColor = 0);
        //! Gibt die Schriftgre zurck, die zum Zeichnen verwendet wird
        DLLEXPORT int getSchriftSize() const;
        //! Gibt den Abstand in Pixeln zum unteren Ende der darber ligenden
        //! Zeile zurck
        DLLEXPORT int getZeilenabstand() const;
        //! Ermittelt, wie viele Pixel bentigt werden, um einen Bestimmten Text
        //! vollstndig darzustellen \param txt Der Text, von dem die Breite in
        //! Pixeln ermitelt werden soll
        DLLEXPORT virtual int getTextBreite(const char* txt) const;
        //! Ermittelt, wie viele Pixel bentigt werden, um einen Bestimmten Text
        //! vollstndig darzustellen \param txt Der Text, von dem die Hhe in
        //! Pixeln ermitelt werden soll
        DLLEXPORT virtual int getTextHeight(const char* txt) const;
        //! Ermittelt, wie viele Pixel bentigt werden, um einen Bestimmten
        //! Buchstaben vollstndig darzustellen \param c Der Buchstabe, von dem
        //! die Breite in Pixeln ermitelt werden soll
        DLLEXPORT virtual int getCharWidth(const char c) const;
        //! Ermittelt, wie viele Pixel bentigt werden, um einen Bestimmten Text
        //! vollstndig darzustellen \param c Der Buchstabe, von dem die Hhe in
        //! Pixeln ermitelt werden soll
        DLLEXPORT virtual int getCharHeight(const char c) const;
        //! Gibt den Abstand in Pixeln zum unteren Ende der darber ligenden
        //! Zeile zurck
        DLLEXPORT int getZeilenAbstand() const;
        //! Gibt den Abstand in Pixeln zum zwischen zwei zeichen auf der x Achse
        //! zurck
        DLLEXPORT int getZeichenAbstand() const;
        //! Gibt die skallierte Hhe zurck, die eine gezeichnete Zeile in
        //! Pixeln bentigt
        DLLEXPORT int getZeilenHeight() const;
        //! Ermittelt das Zeichen im Text, auf das die Maus zeigt
        //! \param zTxt Der Text, auf den die Maus Zeigt
        //! \param mausX Die X Position der Maus in Pixeln Relativ zur Position
        //! des ersten Zeichens \param mausY Die Y Position der Maus in Pixeln
        //! Relativ zur Position des ersten Zeichens
        DLLEXPORT virtual int textPos(
            const char* txt, int mausX, int mausY) const;
    };

    class GravurTextRenderer : public TextRenderer
    {
    public:
        DLLEXPORT GravurTextRenderer();
        DLLEXPORT GravurTextRenderer(Schrift* schrift);
        DLLEXPORT virtual ~GravurTextRenderer();
        //! Zeichnet einen Bestimmten Buchstaben mit einfrbung auf ein Bild
        //! Nutze (setPosition) und (setDrawSchriftGre) um die Position und
        //! die Gre zu verndern \param x x position des ersten zeichens
        //! \param y y position des ersten zeichens
        //! \param txt Der Text, der gezeichnet werden soll
        //! \param zRObj Das Bild, auf das gezeichnet werden soll
        //! \param color Die Farbe, in der der Text gezeichnet werden soll
        //! \param underlined 1, falls der Text unterstrichen sein soll
        //! \param selected 1, falls das zeichen eingefrbt sein soll
        //! \param selectedBackgroundColor Die Hintergrund Farbe des
        //! eingefrbten Textes
        DLLEXPORT virtual void renderChar(int& x,
            int y,
            char c,
            Bild& zRObj,
            int color,
            bool underlined = 0,
            bool selected = 0,
            int selectedBackgroundColor = 0) override;
        //! Ermittelt, wie viele Pixel bentigt werden, um einen Bestimmten
        //! Buchstaben vollstndig darzustellen \param c Der Buchstabe, von dem
        //! die Breite in Pixeln ermitelt werden soll
        DLLEXPORT virtual int getCharWidth(const char c) const override;
        //! Ermittelt, wie viele Pixel bentigt werden, um einen Bestimmten Text
        //! vollstndig darzustellen \param c Der Buchstabe, von dem die Hhe in
        //! Pixeln ermitelt werden soll
        DLLEXPORT virtual int getCharHeight(const char c) const override;
    };

    class KursivTextRenderer : public TextRenderer
    {
    public:
        DLLEXPORT KursivTextRenderer();
        DLLEXPORT KursivTextRenderer(Schrift* schrift);
        DLLEXPORT virtual ~KursivTextRenderer();
        //! Zeichnet einen Bestimmten Buchstaben mit einfrbung auf ein Bild
        //! Nutze (setPosition) und (setDrawSchriftGre) um die Position und
        //! die Gre zu verndern \param x x position des ersten zeichens
        //! \param y y position des ersten zeichens
        //! \param txt Der Text, der gezeichnet werden soll
        //! \param zRObj Das Bild, auf das gezeichnet werden soll
        //! \param color Die Farbe, in der der Text gezeichnet werden soll
        //! \param underlined 1, falls der Text unterstrichen sein soll
        //! \param selected 1, falls das zeichen eingefrbt sein soll
        //! \param selectedBackgroundColor Die Hintergrund Farbe des
        //! eingefrbten Textes
        DLLEXPORT virtual void renderChar(int& x,
            int y,
            char c,
            Bild& zRObj,
            int color,
            bool underlined = 0,
            bool selected = 0,
            int selectedBackgroundColor = 0) override;
        //! Ermittelt, wie viele Pixel bentigt werden, um einen Bestimmten
        //! Buchstaben vollstndig darzustellen \param c Der Buchstabe, von dem
        //! die Breite in Pixeln ermitelt werden soll
        DLLEXPORT virtual int getCharWidth(const char c) const override;
    };
} // namespace Framework

#endif